﻿Imports Infragistics.Controls.Maps

'   Implementation of MapTileSource to fetch map from WMS
'   Mathematics and coordinate calculations are found at http://www.xunpeng.net/dev/java/20111125/167.html

Public Class cMapTileSourceWms
    Inherits MapTileSource

    Private Const TILE_SIZE As Integer = 256
    Private Const EARTH_RADIUS As Double = 6378137
    Private Const EARTH_CIRCUMFERENCE As Double = EARTH_RADIUS * 2 * Math.PI
    Private Const HALF_EARTH_CIRCUMFERENCE As Double = EARTH_CIRCUMFERENCE / 2

#Region "MapTileSource implementation"

    Private m_WmsUrl As String = ""
    Private m_ErrorUrl As String = ""

    Private m_ErrorTile As Integer = 0

    Public Sub New()

        MyBase.New(Integer.MaxValue, Integer.MaxValue, TILE_SIZE, TILE_SIZE, 0)

        m_WmsUrl = "http://www2.demis.nl/WMS/wms.ashx?WMS=SAMPLE&version=1.3.0&request=GetMap&layers=Countries,Coastline,Borders&styles=&crs=CRS:84&bbox={minx},{miny},{maxx},{maxy}&width={width}&height={height}&format=image/png"
        m_ErrorUrl = "http://www2.demis.nl/WMS/wms.ashx?WMS=SAMPLE&version=1.3.0&request=GetMap"

    End Sub

    Protected Overrides Sub GetTileLayers(tileLevel As Integer, tilePositionX As Integer, tilePositionY As Integer, tileImageLayerSources As System.Collections.Generic.IList(Of Object))
        If Not m_WmsUrl = "" Then
            Dim zoom As Integer = tileLevel - 8
            If zoom > 0 Then
                Dim quadKey As String = TileXYToQuadKey(tilePositionX, tilePositionY, zoom)
                Dim boundingBox As cBBox = QuadKeyToBBox(quadKey)

                Dim minX As Double = XToLongitudeAtZoom(boundingBox.X * TILE_SIZE, 18)
                Dim minY As Double = YToLatitudeAtZoom(boundingBox.Y * TILE_SIZE, 18)

                Dim maxX As Double = XToLongitudeAtZoom((boundingBox.X + boundingBox.Width) * TILE_SIZE, 18)
                Dim maxY As Double = YToLatitudeAtZoom((boundingBox.Y - boundingBox.Height) * TILE_SIZE, 18)

                Dim tileW As Double = TILE_SIZE
                Dim tileH As Double = TILE_SIZE

                Dim url As String = m_WmsUrl.Replace("{minx}", minX.ToString.Replace(",", ".")).
                                           Replace("{miny}", minY.ToString.Replace(",", ".")).
                                           Replace("{maxx}", maxX.ToString.Replace(",", ".")).
                                           Replace("{maxy}", maxY.ToString.Replace(",", ".")).
                                           Replace("{width}", tileW.ToString.Replace(",", ".")).
                                           Replace("{height}", tileH.ToString.Replace(",", "."))

                '   Trigger error?
                m_ErrorTile += 1
                If m_ErrorTile Mod 20 = 0 Then url = m_ErrorUrl

                tileImageLayerSources.Add(New Uri(url))
                QueryTile(url)
            End If
        End If
    End Sub

    Private Sub QueryTile(url As String)
        Dim wc As WebClient = New WebClient()
        AddHandler wc.DownloadStringCompleted, AddressOf QueryTileCompleted
        wc.DownloadStringAsync(New Uri(url), url)
    End Sub

    Private Sub QueryTileCompleted(sender As Object, e As DownloadStringCompletedEventArgs)
        System.Diagnostics.Debugger.Log(0, Nothing, Environment.NewLine)
        System.Diagnostics.Debugger.Log(0, Nothing, e.UserState.ToString & Environment.NewLine)
        If e.Error Is Nothing Then
            Dim tile As String = e.Result
            System.Diagnostics.Debugger.Log(0, Nothing, tile & Environment.NewLine)
        Else
            System.Diagnostics.Debugger.Log(0, Nothing, e.Error.ToString & Environment.NewLine)
        End If
    End Sub

#End Region

#Region "Mr. Jeeves"

    Private Function QuadKeyToBBox(quadKey As String, x As Integer, y As Integer, zoomLevel As Integer) As cBBox
        Dim c As Char = quadKey(0)

        Dim tileSize As Integer = 2 << (18 - zoomLevel - 1)

        If c = "0"c Then
            y = y - tileSize

        ElseIf c = "1"c Then
            y = y - tileSize
            x = x + tileSize

        ElseIf c = "3"c Then
            x = x + tileSize
        End If

        If quadKey.Length > 1 Then
            Return QuadKeyToBBox(quadKey.Substring(1), x, y, zoomLevel + 1)
        End If
        Return New cBBox(x, y, tileSize, tileSize)
    End Function

    Private Function QuadKeyToBBox(quadKey As String) As cBBox
        Const x As Integer = 0
        Const y As Integer = 262144
        Return QuadKeyToBBox(quadKey, x, y, 1)
    End Function

    Private Function XToLongitudeAtZoom(x As Integer, zoom As Integer) As Double
        Dim arc As Double = EARTH_CIRCUMFERENCE / ((1 << zoom) * TILE_SIZE)
        Dim metersX As Double = (x * arc) - HALF_EARTH_CIRCUMFERENCE
        Dim result As Double = RadToDeg(metersX / EARTH_RADIUS)
        Return result
    End Function

    Private Function YToLatitudeAtZoom(y As Integer, zoom As Integer) As Double
        Dim arc As Double = EARTH_CIRCUMFERENCE / ((1 << zoom) * TILE_SIZE)
        Dim metersY As Double = HALF_EARTH_CIRCUMFERENCE - (y * arc)
        Dim a As Double = Math.Exp(metersY * 2 / EARTH_RADIUS)
        Dim result As Double = RadToDeg(Math.Asin((a - 1) / (a + 1)))
        Return result
    End Function

    Private Function RadToDeg(d As Double) As Double
        Return d / Math.PI * 180.0
    End Function

    Private Shared Function TileXYToQuadKey(tileX As Integer, tileY As Integer, levelOfDetail As Integer) As String
        Dim quadKey = New Text.StringBuilder()
        For i As Integer = levelOfDetail To 1 Step -1
            Dim digit As Char = "0"c
            Dim mask As Integer = 1 << (i - 1)
            If (tileX And mask) <> 0 Then
                digit = IncrementChar(digit)
            End If
            If (tileY And mask) <> 0 Then
                digit = IncrementChar(digit)
                digit = IncrementChar(digit)
            End If
            quadKey.Append(digit)
        Next
        Return quadKey.ToString()
    End Function

    Private Shared Function IncrementChar(c As Char) As Char
        Dim a() As Byte = System.Text.Encoding.UTF8.GetBytes(c)
        a(0) += 1
        Return System.Text.Encoding.UTF8.GetChars(a).First
    End Function

#End Region

End Class

#Region "Helper classes"

Public Class cBBox

    Public Property X As Integer
    Public Property Y As Integer
    Public Property Width As Integer
    Public Property Height As Integer

    Public Sub New(_x As Integer, _y As Integer, _w As Integer, _h As Integer)
        X = _x
        Y = _y
        Width = _w
        Height = _h
    End Sub

End Class

#End Region
